@@ -1,18 +1,21 @@
 # pylint:disable=too-many-boolean-expressions
 import logging
-from typing import Optional, TYPE_CHECKING
+from typing import Optional, Union, Type, Iterable, Tuple, TYPE_CHECKING
 
-from ailment.statement import Assignment, ConditionalJump, Call
-from ailment.expression import Expression, Convert, Tmp, Register, Load, BinaryOp, UnaryOp, Const, ITE
+from ailment.statement import Statement, Assignment, Call
+from ailment.expression import Expression, Tmp, Register, Load
 
 from ...engines.light.data import SpOffset
 from ...knowledge_plugins.key_definitions.constants import OP_AFTER
 from ...knowledge_plugins.key_definitions import atoms
 from ...analyses.reaching_definitions.external_codeloc import ExternalCodeLocation
-
 from .. import Analysis, register_analysis
+from .peephole_optimizations import STMT_OPTS, EXPR_OPTS, PeepholeOptimizationStmtBase, PeepholeOptimizationExprBase
+from .ailblock_walker import AILBlockWalker
 
 if TYPE_CHECKING:
+    from angr.storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
+    from angr.knowledge_plugins.key_definitions.live_definitions import LiveDefinitions
     from ailment.block import Block
 
 
@@ -23,7 +26,9 @@ class BlockSimplifier(Analysis):
     """
     Simplify an AIL block.
     """
-    def __init__(self, block: Optional['Block'], remove_dead_memdefs=False, stack_pointer_tracker=None):
+    def __init__(self, block: Optional['Block'], remove_dead_memdefs=False, stack_pointer_tracker=None,
+                 peephole_optimizations: Optional[Iterable[Union[Type[PeepholeOptimizationStmtBase],Type[PeepholeOptimizationExprBase]]]]=None,
+                 ):
         """
         :param block:   The AIL block to simplify. Setting it to None to skip calling self._analyze(), which is useful
                         in test cases.
@@ -34,6 +39,15 @@ class BlockSimplifier(Analysis):
         self._remove_dead_memdefs = remove_dead_memdefs
         self._stack_pointer_tracker = stack_pointer_tracker
 
+        if peephole_optimizations is None:
+            self._expr_peephole_opts = [ cls(self.project) for cls in EXPR_OPTS ]
+            self._stmt_peephole_opts = [ cls(self.project) for cls in STMT_OPTS ]
+        else:
+            self._expr_peephole_opts = [ cls(self.project) for cls in peephole_optimizations
+                                         if issubclass(cls, PeepholeOptimizationExprBase) ]
+            self._stmt_peephole_opts = [ cls(self.project) for cls in peephole_optimizations
+                                         if issubclass(cls, PeepholeOptimizationStmtBase) ]
+
         self.result_block = None
 
         if self.block is not None:
@@ -69,17 +83,20 @@ class BlockSimplifier(Analysis):
         # propagator
         propagator = self.project.analyses.Propagator(block=block, stack_pointer_tracker=self._stack_pointer_tracker)
         replacements = list(propagator._states.values())[0]._replacements
-        if not replacements:
-            return block
-        new_block = self._replace_and_build(block, replacements)
+        if replacements:
+            _, new_block = self._replace_and_build(block, replacements)
+            new_block = self._eliminate_self_assignments(new_block)
+        else:
+            new_block = block
         new_block = self._eliminate_dead_assignments(new_block)
         new_block = self._peephole_optimize(new_block)
         return new_block
 
     @staticmethod
-    def _replace_and_build(block, replacements):
+    def _replace_and_build(block, replacements) -> Tuple[bool,'Block']:
 
         new_statements = block.statements[::]
+        replaced = False
 
         for codeloc, repls in replacements.items():
             for old, new in repls.items():
@@ -97,15 +114,25 @@ class BlockSimplifier(Analysis):
                         # special case: do not replace the ret_expr of a call statement to another call statement
                         r = False
                         new_stmt = None
+                    elif isinstance(stmt, Assignment):
+                        # special case: do not replace the dst
+                        new_stmt = None
+                        r, new_src = stmt.src.replace(old, new)
+                        if r:
+                            new_stmt = Assignment(stmt.idx, stmt.dst, new_src, **stmt.tags)
                     else:
                         r, new_stmt = stmt.replace(old, new)
 
                 if r:
+                    replaced = True
                     new_statements[codeloc.stmt_idx] = new_stmt
 
+        if not replaced:
+            return False, block
+
         new_block = block.copy()
         new_block.statements = new_statements
-        return new_block
+        return True, new_block
 
     @staticmethod
     def _eliminate_self_assignments(block):
@@ -133,7 +160,7 @@ class BlockSimplifier(Analysis):
                                                        )
 
         used_tmp_indices = set(rd.one_result.tmp_uses.keys())
-        live_defs = rd.one_result
+        live_defs: 'LiveDefinitions' = rd.one_result
 
         # Find dead assignments
         dead_defs_stmt_idx = set()
@@ -154,12 +181,21 @@ class BlockSimplifier(Analysis):
                     # is entirely possible that at the end of the block, a register definition is not used.
                     # however, it might be used in future blocks.
                     # so we only remove a definition if the definition is not alive anymore at the end of the block
+                    defs_ = set()
                     if isinstance(d.atom, atoms.Register):
-                        if d not in live_defs.register_definitions.get_variables_by_offset(d.atom.reg_offset):
-                            dead_defs_stmt_idx.add(d.codeloc.stmt_idx)
-                    if isinstance(d.atom, atoms.MemoryLocation) and isinstance(d.atom.addr, SpOffset):
-                        if d not in live_defs.stack_definitions.get_variables_by_offset(d.atom.addr.offset):
-                            dead_defs_stmt_idx.add(d.codeloc.stmt_idx)
+                        vs: 'MultiValues' = live_defs.register_definitions.load(d.atom.reg_offset, size=d.atom.size)
+                    elif isinstance(d.atom, atoms.MemoryLocation) and isinstance(d.atom.addr, SpOffset):
+                        stack_addr = live_defs.stack_offset_to_stack_addr(d.atom.addr.offset)
+                        vs: 'MultiValues' = live_defs.stack_definitions.load(stack_addr, size=d.atom.size,
+                                                                             endness=d.atom.endness)
+                    else:
+                        continue
+                    for values in vs.values.values():
+                        for value in values:
+                            defs_.update(live_defs.extract_defs(value))
+
+                    if d not in defs_:
+                        dead_defs_stmt_idx.add(d.codeloc.stmt_idx)
 
         # Remove dead assignments
         for idx, stmt in enumerate(block.statements):
@@ -190,96 +226,79 @@ class BlockSimplifier(Analysis):
 
     def _peephole_optimize(self, block):
 
-        statements = [ ]
-        any_update = False
-        for stmt in block.statements:
-            new_stmt = None
-            if isinstance(stmt, Assignment):
-                new_stmt = self._peephole_optimize_ConstantDereference(stmt)
+        # expressions are updated in place
+        self._peephole_optimize_exprs(block, self._expr_peephole_opts)
 
-            elif isinstance(stmt, ConditionalJump):
-                new_stmt = self._peephole_optimize_ConditionalJump(stmt)
+        statements, stmts_updated = self._peephole_optimize_stmts(block, self._stmt_peephole_opts)
 
-            if new_stmt is not None and new_stmt is not stmt:
-                statements.append(new_stmt)
-                any_update = True
-                continue
-
-            statements.append(stmt)
-
-        if not any_update:
+        if not stmts_updated:
             return block
         new_block = block.copy(statements=statements)
         return new_block
 
-    def _peephole_optimize_ConstantDereference(self, stmt: Assignment):
-        if isinstance(stmt.src, Load) and isinstance(stmt.src.addr, Const):
-            # is it loading from a read-only section?
-            sec = self.project.loader.find_section_containing(stmt.src.addr.value)
-            if sec is not None and sec.is_readable and not sec.is_writable:
-                # do we know the value that it's reading?
-                try:
-                    val = self.project.loader.memory.unpack_word(stmt.src.addr.value, size=self.project.arch.bytes)
-                except KeyError:
-                    return stmt
+    @staticmethod
+    def _peephole_optimize_exprs(block, expr_opts):
 
-                return Assignment(stmt.idx, stmt.dst,
-                                  Const(None, None, val, stmt.src.size * self.project.arch.byte_width),
-                                  **stmt.tags,
-                                  )
+        class _any_update:
+            v = False
 
-        return stmt
+        def _handle_expr(expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement, block) -> Optional[Expression]:
 
-    def _peephole_optimize_ConditionalJump(self, stmt: ConditionalJump):
+            old_expr = expr
 
-        new_condition = self._peephole_optimize_Expr(stmt.condition)
+            redo = True
+            while redo:
+                redo = False
+                for expr_opt in expr_opts:
+                    if isinstance(expr, expr_opt.expr_classes):
+                        r = expr_opt.optimize(expr)
+                        if r is not None and r is not expr:
+                            expr = r
+                            redo = True
+                            break
 
-        # if (!cond) {} else { ITE(cond, true_branch, false_branch } ==> if (cond) { ITE(...) } else {}
-        if isinstance(stmt.false_target, ITE) and \
-                isinstance(new_condition, UnaryOp) and \
-                new_condition.op == "Not":
-            new_true_target = stmt.false_target
-            new_false_target = stmt.true_target
-            new_condition = new_condition.operand
-        else:
-            new_true_target = stmt.true_target
-            new_false_target = stmt.false_target
+            if expr is not old_expr:
+                _any_update.v = True
+                # continue to process the expr
+                r = AILBlockWalker._handle_expr(walker, expr_idx, expr, stmt_idx, stmt, block)
+                return expr if r is None else r
 
-        if new_condition is not stmt.condition or \
-                new_true_target is not stmt.true_target or \
-                new_false_target is not stmt.false_target:
-            # it's updated
-            return self._peephole_optimize_ConditionalJump(
-                ConditionalJump(stmt.idx, new_condition, new_true_target, new_false_target, **stmt.tags)
-            )
+            return AILBlockWalker._handle_expr(walker, expr_idx, expr, stmt_idx, stmt, block)
 
-        # if (cond) {ITE(cond, true_branch, false_branch)} else {} ==> if (cond) {true_branch} else {}
-        if isinstance(stmt.true_target, ITE) and new_condition == stmt.true_target.cond:
-            new_true_target = stmt.true_target.iftrue
-        else:
-            new_true_target = stmt.true_target
+        # run expression optimizers
+        walker = AILBlockWalker()
+        walker._handle_expr = _handle_expr
+        walker.walk(block)
 
-        if new_condition is not stmt.condition or new_true_target is not stmt.true_target:
-            # it's updated
-            return self._peephole_optimize_ConditionalJump(
-                ConditionalJump(stmt.idx, new_condition, new_true_target, stmt.false_target, **stmt.tags)
-            )
+        return _any_update.v
 
-        return stmt
+    @staticmethod
+    def _peephole_optimize_stmts(block, stmt_opts):
 
-    def _peephole_optimize_Expr(self, expr: Expression):
+        any_update = False
+        statements = [ ]
 
-        # Convert(N->1, (Convert(1->N, t_x) ^ 0x1<N>) ==> Not(t_x)
-        if isinstance(expr, Convert) and \
-                isinstance(expr.operand, BinaryOp) and \
-                expr.operand.op == "Xor" and \
-                isinstance(expr.operand.operands[0], Convert) and \
-                isinstance(expr.operand.operands[1], Const) and \
-                expr.operand.operands[1].value == 1:
-            new_expr = UnaryOp(None, "Not", expr.operand.operands[0].operand)
-            return self._peephole_optimize_Expr(new_expr)
+        # run statement optimizers
+        for stmt in block.statements:
+            old_stmt = stmt
+            redo = True
+            while redo:
+                redo = False
+                for opt in stmt_opts:
+                    if isinstance(stmt, opt.stmt_classes):
+                        r = opt.optimize(stmt)
+                        if r is not None and r is not stmt:
+                            stmt = r
+                            redo = True
+                            break
 
-        return expr
+            if stmt is not None and stmt is not old_stmt:
+                statements.append(stmt)
+                any_update = True
+            else:
+                statements.append(old_stmt)
+
+        return statements, any_update
 
 
 register_analysis(BlockSimplifier, 'AILBlockSimplifier')
