@@ -12,7 +12,7 @@ from ...utils.graph import dominates, shallow_reverse
 from ...block import Block, BlockNode
 from ..cfg.cfg_utils import CFGUtils
 from .structurer_nodes import (EmptyBlockNotice, SequenceNode, CodeNode, SwitchCaseNode, BreakNode,
-                               ConditionalBreakNode, LoopNode, ConditionNode, ContinueNode)
+                               ConditionalBreakNode, LoopNode, ConditionNode, ContinueNode, CascadingConditionNode)
 from .utils import extract_jump_targets, switch_extract_cmp_bounds
 
 l = logging.getLogger(__name__)
@@ -276,6 +276,20 @@ class ConditionProcessor:
                                  self.remove_claripy_bool_asts(node.false_node, memo=memo),
                                  )
 
+        elif isinstance(node, CascadingConditionNode):
+
+            cond_and_nodes = [ ]
+            for cond, child_node in node.condition_and_nodes:
+                cond_and_nodes.append((
+                    self.convert_claripy_bool_ast(cond, memo=memo),
+                    self.remove_claripy_bool_asts(child_node, memo=memo))
+                )
+            else_node = None if node.else_node is None else self.remove_claripy_bool_asts(node.else_node, memo=memo)
+            return CascadingConditionNode(node.addr,
+                                          cond_and_nodes,
+                                          else_node=else_node,
+                                          )
+
         elif isinstance(node, LoopNode):
 
             result = node.copy()
@@ -337,6 +351,16 @@ class ConditionProcessor:
             if s is None and block.false_node:
                 s = cls.get_last_statement(block.false_node)
             return s
+        if type(block) is CascadingConditionNode:
+            s = None
+            if block.else_node is not None:
+                s = cls.get_last_statement(block.else_node)
+            else:
+                for _, node in reversed(block.condition_and_nodes):
+                    s = cls.get_last_statement(node)
+                    if s is not None:
+                        break
+            return s
         if type(block) is BreakNode:
             return None
         if type(block) is ContinueNode:
@@ -397,6 +421,18 @@ class ConditionProcessor:
                 last_stmts = cls.get_last_statements(block.false_node)
                 s.extend(last_stmts)
             return s
+        if type(block) is CascadingConditionNode:
+            s = [ ]
+            if block.else_node is not None:
+                try:
+                    last_stmts = cls.get_last_statements(block.else_node)
+                    s.extend(last_stmts)
+                except EmptyBlockNotice:
+                    pass
+            for _, node in block.condition_and_nodes:
+                last_stmts = cls.get_last_statements(node)
+                s.extend(last_stmts)
+            return s
         if type(block) is BreakNode:
             return [ block ]
         if type(block) is ContinueNode:
@@ -488,10 +524,10 @@ class ConditionProcessor:
 
         if memo is None:
             memo = {}
-        if cond in memo:
-            return memo[cond]
+        if cond._hash in memo:
+            return memo[cond._hash]
         r = self.convert_claripy_bool_ast_core(cond, memo)
-        memo[cond] = r
+        memo[cond._hash] = r
         return r
 
     def convert_claripy_bool_ast_core(self, cond, memo):
@@ -584,6 +620,11 @@ class ConditionProcessor:
             operand1 = ailment.Expr.Convert(None, operand1.bits, operand0.bits, False, operand1)
             return op(conv(operand0), conv(operand1))
 
+        def _dummy_bvs(condition):
+            var = claripy.BVS('ailexpr_%s' % repr(condition), condition.bits, explicit_name=True)
+            self._condition_mapping[var.args[0]] = condition
+            return var
+
         _mapping = {
             'LogicalAnd': lambda expr, conv: claripy.And(conv(expr.operands[0]), conv(expr.operands[1])),
             'LogicalOr': lambda expr, conv: claripy.Or(conv(expr.operands[0]), conv(expr.operands[1])),
@@ -600,6 +641,7 @@ class ConditionProcessor:
             'Add': lambda expr, conv: conv(expr.operands[0]) + conv(expr.operands[1]),
             'Sub': lambda expr, conv: conv(expr.operands[0]) - conv(expr.operands[1]),
             'Mul': lambda expr, conv: conv(expr.operands[0]) * conv(expr.operands[1]),
+            'Div': lambda expr, conv: conv(expr.operands[0]) / conv(expr.operands[1]),
             'Not': lambda expr, conv: claripy.Not(conv(expr.operand)),
             'Xor': lambda expr, conv: conv(expr.operands[0]) ^ conv(expr.operands[1]),
             'And': lambda expr, conv: conv(expr.operands[0]) & conv(expr.operands[1]),
@@ -607,13 +649,15 @@ class ConditionProcessor:
             'Shr': lambda expr, conv: _op_with_unified_size(claripy.LShR, conv, expr.operands[0], expr.operands[1]),
             'Shl': lambda expr, conv: _op_with_unified_size(operator.lshift, conv, expr.operands[0], expr.operands[1]),
             'Sar': lambda expr, conv: _op_with_unified_size(operator.rshift, conv, expr.operands[0], expr.operands[1]),
+
+            # There is no claripy operation for the following operations
+            'DivMod': lambda expr, _: _dummy_bvs(expr),
+            'CmpF': lambda expr, _: _dummy_bvs(expr),
         }
 
         if isinstance(condition, (ailment.Expr.Load, ailment.Expr.DirtyExpression, ailment.Expr.BasePointerOffset,
                                   ailment.Expr.ITE, ailment.Stmt.Call)):
-            var = claripy.BVS('ailexpr_%s' % repr(condition), condition.bits, explicit_name=True)
-            self._condition_mapping[var.args[0]] = condition
-            return var
+            return _dummy_bvs(condition)
         elif isinstance(condition, ailment.Expr.Register):
             var = claripy.BVS('ailexpr_%s-%d' % (repr(condition), condition.idx), condition.bits, explicit_name=True)
             self._condition_mapping[var.args[0]] = condition
