@@ -86,7 +86,7 @@ class SimEngineVRBase(SimEngineLight):
     def _reference(self, richr: RichR, codeloc: CodeLocation, src=None):
         data: claripy.ast.Base = richr.data
         # extract stack offset
-        if self.state.is_stack_address(data):
+        if data is not None and self.state.is_stack_address(data):
             # this is a stack address
             stack_offset: Optional[int] = self.state.get_stack_offset(data)
         else:
@@ -100,7 +100,7 @@ class SimEngineVRBase(SimEngineLight):
         # find the correct variable
         existing_vars: List[Tuple[SimVariable,int]] = [ ]
         for candidate, offset in var_candidates:
-            if isinstance(candidate, SimStackVariable) and candidate.offset == -stack_offset:
+            if isinstance(candidate, SimStackVariable) and candidate.offset == stack_offset:
                 existing_vars.append((candidate, offset))
         variable = None
         if existing_vars:
@@ -314,13 +314,19 @@ class SimEngineVRBase(SimEngineLight):
                                        endness=self.state.arch.memory_endness if stmt is None else stmt.endness)
 
         codeloc = CodeLocation(self.block.addr, self.stmt_idx, ins_addr=self.ins_addr)
-        values: MultiValues = self.state.global_region.load(addr,
-                                                            size=size,
-                                                            endness=self.state.arch.memory_endness if stmt is None else stmt.endness)
-        for vs in values.values.values():
-            for v in vs:
-                for var_offset, var in self.state.extract_variables(v):
-                    variable_manager.write_to(var, var_offset, codeloc, atom=stmt)
+        try:
+            values: MultiValues = self.state.global_region.load(
+                addr,
+                size=size,
+                endness=self.state.arch.memory_endness if stmt is None else stmt.endness)
+        except SimMemoryMissingError:
+            values = None
+
+        if values is not None:
+            for vs in values.values.values():
+                for v in vs:
+                    for var_offset, var in self.state.extract_variables(v):
+                        variable_manager.write_to(var, var_offset, codeloc, atom=stmt)
 
         # create type constraints
         if data.typevar is not None:
@@ -485,7 +491,7 @@ class SimEngineVRBase(SimEngineLight):
             addr_v: int = addr._model_concrete.value
             variables = global_variables.get_global_variables(addr_v)
             if not variables:
-                var = SimMemoryVariable(addr_v, size)
+                var = SimMemoryVariable(addr_v, size, ident=global_variables.next_variable_ident('global'))
                 global_variables.add_variable('global', addr_v, var)
                 variables = [var]
             for var in variables:
@@ -579,4 +585,8 @@ class SimEngineVRBase(SimEngineLight):
                     # typevar = next(reversed(list(self.state.typevars[var].values())))
                     typevar = self.state.typevars[var]
 
-        return RichR(next(iter(value_list[0])), variable=var, typevar=typevar)
+        if len(value_list) == 1:
+            r_value = next(iter(value_list[0]))
+        else:
+            r_value = self.state.top(size * self.arch.byte_width)  # fall back to top
+        return RichR(r_value, variable=var, typevar=typevar)
