@@ -1,5 +1,6 @@
 # pylint:disable=no-self-use,isinstance-second-argument-not-valid-type
 from typing import Tuple, Optional, Union, Any
+import struct
 import logging
 
 import ailment
@@ -8,7 +9,7 @@ import claripy
 import archinfo
 
 from ...engines.vex.claripy.datalayer import value as claripy_value
-from ...engines.vex.claripy.irop import UnsupportedIROpError, vexop_to_simop
+from ...engines.vex.claripy.irop import UnsupportedIROpError, SimOperationError, vexop_to_simop
 from ...code_location import CodeLocation
 from ...utils.constants import DEFAULT_STATEMENT
 from ..engine import SimEngine
@@ -268,7 +269,7 @@ class SimEngineLightVEXMixin(SimEngineLightMixin):
         simop = None
         try:
             simop = vexop_to_simop(expr.op)
-        except UnsupportedIROpError:
+        except (UnsupportedIROpError, SimOperationError):
             pass
 
         if simop is not None and simop.op_attrs.get('conversion', None):
@@ -789,6 +790,22 @@ class SimEngineLightAILMixin(SimEngineLightMixin):
     def _ail_handle_CallExpr(self, expr):
         raise NotImplementedError('Please implement the CallExpr handler with your own logic.')
 
+    def _ail_handle_Reinterpret(self, expr: ailment.Expr.Reinterpret):
+        arg = self._expr(expr.operand)
+
+        if isinstance(arg, int) and expr.from_bits == 32 and expr.from_type == "I" and expr.to_bits == 32 and expr.to_type == "F":
+            # int -> float
+            b = struct.pack("<I", arg)
+            f = struct.unpack("<f", b)[0]
+            return f
+        elif isinstance(arg, float) and expr.from_bits == 32 and expr.from_type == "F" and expr.to_bits == 32 and expr.to_type == "I":
+            # float -> int
+            b = struct.pack("<f", arg)
+            v = struct.unpack("<I", b)[0]
+            return v
+
+        return expr
+
     def _ail_handle_UnaryOp(self, expr):
         handler_name = '_ail_handle_%s' % expr.op
         try:
@@ -903,10 +920,24 @@ class SimEngineLightAILMixin(SimEngineLightMixin):
         try:
             return expr_0 * expr_1
         except TypeError:
-            return ailment.Expr.BinaryOp(expr.idx, 'Mul', [expr_0, expr_1], expr.signed, **expr.tags)
+            return ailment.Expr.BinaryOp(expr.idx, 'Mul', [expr_0, expr_1], expr.signed, bits=expr.bits, **expr.tags)
 
     def _ail_handle_Mull(self, expr):
-        return self._ail_handle_Mul(expr)
+
+        arg0, arg1 = expr.operands
+
+        expr_0 = self._expr(arg0)
+        expr_1 = self._expr(arg1)
+
+        if expr_0 is None:
+            expr_0 = arg0
+        if expr_1 is None:
+            expr_1 = arg1
+
+        try:
+            return expr_0 * expr_1
+        except TypeError:
+            return ailment.Expr.BinaryOp(expr.idx, 'Mull', [expr_0, expr_1], expr.signed, bits=expr.bits, **expr.tags)
 
     def _ail_handle_And(self, expr):
 
