@@ -137,6 +137,8 @@ class Tracer(ExplorationTechnique):
     :param mode:                Tracing mode.
     :param aslr:                Whether there are aslr slides. if not, tracer uses trace address
                                 as state address.
+    :param follow_unsat:        Whether unsatisfiable states should be treated as potential
+                                successors or not.
 
     :ivar predecessors:         A list of states in the history before the final state.
     """
@@ -149,7 +151,8 @@ class Tracer(ExplorationTechnique):
             copy_states=False,
             fast_forward_to_entry=True,
             mode=TracingMode.Strict,
-            aslr=True):
+            aslr=True,
+            follow_unsat=False):
         super().__init__()
         self._trace = trace
         self._resiliency = resiliency
@@ -157,6 +160,7 @@ class Tracer(ExplorationTechnique):
         self._copy_states = copy_states
         self._mode = mode
         self._aslr = aslr
+        self._follow_unsat = follow_unsat
         self._fast_forward_to_entry = fast_forward_to_entry
 
         self._aslr_slides = {}  # type: Dict[cle.Backend, int]
@@ -186,27 +190,29 @@ class Tracer(ExplorationTechnique):
             indices = set(i for i in indices if self._filter_idx(angr_addr, i))
             threshold //= 2
 
-
         return indices
 
-    def _filter_idx(self, angr_addr, idx):
-        slide = self._trace[idx] - angr_addr
-        block = self.project.factory.block(angr_addr)
-        legal_next = block.vex.constant_jump_targets
-        return not legal_next or any(a + slide == self._trace[idx + 1] for a in legal_next)
-
-    def setup(self, simgr):
-        simgr.populate('missed', [])
-        simgr.populate('traced', [])
-        simgr.populate('crashed', [])
-        simgr.populate('desync', [])
-
-        if len(simgr.active) != 1:
-            raise AngrTracerError("Tracer is being invoked on a SimulationManager without exactly one active state")
-
-        # handle aslr slides
+    def _identify_aslr_slides(self):
+        """
+        libraries can be mapped differently in the original run(in the trace) and in angr
+        this function identifies the difference(called aslr slides) of each library to help angr translate
+        original address and address in angr back and forth
+        """
         if self._aslr:
+            # if we don't know whether there is any slide, we need to identify the slides via heuristics
             for obj in self.project.loader.all_elf_objects:
+                # heuristic 1: non-PIC  objects are loaded without aslr slides
+                if not obj.pic:
+                    self._aslr_slides[obj] = 0
+                    continue
+
+                # heuristic 2: library objects with custom_base_addr are loaded at the correct locations
+                if obj._custom_base_addr:
+                    l.info("%s is assumed to be loaded at the address matching the one in the trace", obj)
+                    self._aslr_slides[obj] = 0
+                    continue
+
+                # heuristic 3: entry point of an object should appear in the trace
                 possibilities = None
                 for entry in obj.initializers + ([obj.entry] if obj.is_main_bin else []):
                     indices = self._locate_entry_point(entry)
@@ -225,12 +231,35 @@ class Tracer(ExplorationTechnique):
                     self._aslr_slides[obj] = next(iter(possibilities))
                 else:
                     raise AngrTracerError("Trace seems ambiguous with respect to what the ASLR slides are for %s. This is surmountable, please open an issue." % obj)
-
         else:
+            # if we know there is no slides, just trust the address in the loader
             for obj in self.project.loader.all_elf_objects:
                 self._aslr_slides[obj] = 0
             self._current_slide = 0
 
+    def _filter_idx(self, angr_addr, idx):
+        slide = self._trace[idx] - angr_addr
+        block = self.project.factory.block(angr_addr)
+        legal_next = block.vex.constant_jump_targets
+        if legal_next:
+            return any(a + slide == self._trace[idx + 1] for a in legal_next)
+        else:
+            # the intuition is that if the first block of an initializer does an indirect jump,
+            # it's probably a call out to another binary (notably __libc_start_main)
+            # this is an awful fucking heuristic but it's as good as we've got
+            return abs(self._trace[idx] - self._trace[idx + 1]) > 0x1000
+
+    def setup(self, simgr):
+        simgr.populate('missed', [])
+        simgr.populate('traced', [])
+        simgr.populate('crashed', [])
+        simgr.populate('desync', [])
+
+        if len(simgr.active) != 1:
+            raise AngrTracerError("Tracer is being invoked on a SimulationManager without exactly one active state")
+
+        self._identify_aslr_slides()
+
         if self._fast_forward_to_entry:
             idx = self._trace.index(self._translate_state_addr(self.project.entry))
             # step to entry point
@@ -314,8 +343,8 @@ class Tracer(ExplorationTechnique):
         sat_succs = succs_dict[None]  # satisfiable states
         succs = sat_succs + succs_dict['unsat']  # both satisfiable and unsatisfiable states
 
-        if self._mode == TracingMode.Permissive:
-            # permissive mode
+        if not self._follow_unsat:
+            # Only satisfiable states need to be checked for correct successor
             if len(sat_succs) == 1:
                 try:
                     self._update_state_tracking(sat_succs[0])
@@ -331,7 +360,7 @@ class Tracer(ExplorationTechnique):
                 succs_dict[None] = [succ]
                 succs_dict['missed'] = [s for s in sat_succs if s is not succ]
         else:
-            # strict mode
+            # Check all states for correct successor
             if len(succs) == 1:
                 self._update_state_tracking(succs[0])
             elif len(succs) == 0:
@@ -445,17 +474,30 @@ class Tracer(ExplorationTechnique):
             # ^ this means we will see desyncs of the form unicorn suddenly skips a bunch of qemu blocks
             assert state.history.recent_block_count == len(state.history.recent_bbl_addrs)
 
-            if sync is not None:
-                raise Exception("TODO")
-
-            for addr in state.history.recent_bbl_addrs:
+            for addr_idx, addr in enumerate(state.history.recent_bbl_addrs):
                 if addr == state.unicorn.transmit_addr:
                     continue
 
+                if sync is not None and sync != 'entry':
+                    if self._compare_addr(self._trace[sync], addr):
+                        # Found the address in trace. Start normal trace checks from next address
+                        idx = sync + 1
+                        state.globals['sync_idx'] = None
+                        sync = None
 
-                if self._compare_addr(self._trace[idx], addr) or self._check_qemu_unicorn_desync(state, idx, addr):
+                    continue
+
+                if self._compare_addr(self._trace[idx], addr) or self._check_qemu_unicorn_large_block_split(state, idx, addr_idx):
                     idx += 1
                 else:
+                    is_contained, increment = self._check_qemu_block_in_unicorn_block(state, idx, addr_idx)
+                    if is_contained:
+                        idx += increment
+                        # Big block is now skipped in qemu trace. Perform compare at correct index again.
+                        if self._compare_addr(self._trace[idx], addr):
+                            idx += 1
+                            continue
+
                     raise TracerDesyncError('Oops! angr did not follow the trace',
                                             deviating_addr=addr,
                                             deviating_trace_idx=idx)
@@ -482,6 +524,9 @@ class Tracer(ExplorationTechnique):
         elif state.history.jumpkind.startswith('Ijk_Exit'):
             # termination! will be handled by filter
             pass
+        elif self.project.is_hooked(state.addr) and not self.project.loader.extern_object.contains_addr(state.addr):
+           # handle simprocedures
+           self._sync_return(state, idx)
         elif self._compare_addr(self._trace[idx + 1], state.addr):
             # normal case
             state.globals['trace_idx'] = idx + 1
@@ -577,7 +622,48 @@ class Tracer(ExplorationTechnique):
         else:
             raise AngrTracerError("Trace desynced on jumping into %#x, where no library is mapped!" % state_addr)
 
-    def _check_qemu_unicorn_desync(self, state: 'angr.SimState', trace_curr_idx, state_desync_block_addr):
+    def _check_qemu_block_in_unicorn_block(self, state: 'angr.SimState', trace_curr_idx, state_desync_block_idx):
+        """
+        Check if desync occurred because unicorn block was split into multiple blocks in qemu tracer. If yes, find the
+        correct increment for trace index
+        """
+
+        # We first find the block address where the trace and state's history match
+        for trace_match_idx in range(trace_curr_idx - 1, -1, -1):
+            if self._trace[trace_match_idx] == state.history.recent_bbl_addrs[state_desync_block_idx - 1]:
+                break
+        else:
+            # Failed to find matching block address. qemu block is probably not contained in a previous block.
+            return (False, -1)
+
+        control_flow_insn_types = [CS_GRP_CALL, CS_GRP_IRET, CS_GRP_JUMP, CS_GRP_RET]
+        big_block_start = self._trace[trace_match_idx]
+        big_block_end = None
+        curr_block_addr = big_block_start
+        while True:
+            curr_block = state.project.factory.block(self._translate_trace_addr(curr_block_addr))
+            curr_block_last_insn = curr_block.capstone.insns[-1]
+            if any((curr_block_last_insn.group(insn_type) for insn_type in control_flow_insn_types)):
+                # Found last block
+                big_block_end = curr_block.addr + curr_block.size - 1
+                break
+
+            curr_block_addr = curr_block.addr + curr_block.size
+
+        for last_contain_index in range(trace_match_idx, trace_curr_idx + 1):
+            if self._trace[last_contain_index] < big_block_start or self._trace[last_contain_index] > big_block_end:
+                # This qemu block is not contained in the bigger block
+                return (False, -1)
+
+        # Check for future blocks in trace contained in big block
+        for next_contain_index in range(trace_curr_idx + 1, len(self._trace)):
+            if self._trace[next_contain_index] < big_block_start or self._trace[next_contain_index] > big_block_end:
+                # This qemu block is not contained in bigger block
+                break
+
+        return (True, next_contain_index - trace_curr_idx)
+
+    def _check_qemu_unicorn_large_block_split(self, state: 'angr.SimState', trace_curr_idx, state_desync_block_idx):
         """
         Check if desync occurred because large blocks are split up at different instructions by qemu and unicorn. This
         is done by reconstructing part of block executed so far from the trace and state history and checking if they
@@ -589,7 +675,7 @@ class Tracer(ExplorationTechnique):
         prev_trace_block = state.project.factory.block(self._translate_trace_addr(self._trace[trace_curr_idx - 1]))
         for insn_type in control_flow_insn_types:
             if prev_trace_block.capstone.insns[-1].group(insn_type):
-                # Previous block ends in a control flow instruction. It is a trace desync.
+                # Previous block ends in a control flow instruction. It is not large block different split.
                 return False
 
         # The previous block did not end in a control flow instruction. Let's find the start of this big block from
@@ -612,17 +698,7 @@ class Tracer(ExplorationTechnique):
         # Now we check the part of the state history corresponding to this big basic block to ensure there are no
         # control flow instructions at end of any blocks in the part. This check moves backwards starting from the
         # desyncing block to the start of the big block we found earlier
-        state_history_desync_block_found = False
-        for state_history_block_addr in reversed(state.history.bbl_addrs):
-            if not state_history_desync_block_found:
-                if state_history_block_addr == state_desync_block_addr:
-                    # Found the desync block in state history
-                    state_history_desync_block_found = True
-
-                # Either we haven't found the desync block in state history or we just found it. Since we want to check
-                # for control flow instructions from only from block before the desync block, we can continue here
-                continue
-
+        for state_history_block_addr in reversed(state.history.recent_bbl_addrs[:state_desync_block_idx]):
             state_history_block = state.project.factory.block(state_history_block_addr)
             state_history_block_last_insn = state_history_block.capstone.insns[-1]
             for insn_type in control_flow_insn_types:
@@ -632,26 +708,42 @@ class Tracer(ExplorationTechnique):
                     return False
 
             if state_history_block_addr == big_block_start_addr:
-                if not state_history_desync_block_found:
-                    # We somehow found the start of the big block in the state history but not the block where the
-                    # desync error happened. Something is seriously wrong!
-                    return False
-                else:
-                    # We found start of the big block and no control flow statements in between that and the block where
-                    # desync happend. All good.
-                    break
+                # We found start of the big block and no control flow statements in between that and the block where
+                # desync happend.
+                break
+
+        # Let's find the address of the last byte of the big basic block using VEX lifter
+        angr_big_block_end_addr = None
+        curr_block_addr = big_block_start_addr
+        while True:
+            curr_block = state.project.factory.block(self._translate_trace_addr(curr_block_addr))
+            curr_block_last_insn = curr_block.capstone.insns[-1]
+            if any((curr_block_last_insn.group(insn_type) for insn_type in control_flow_insn_types)):
+                # Found last block
+                angr_big_block_end_addr = curr_block.addr + curr_block.size - 1
+                break
+
+            curr_block_addr = curr_block.addr + curr_block.size
 
         # Let's find the address of the last bytes of the big basic block from the trace
         big_block_end_addr = None
-        for trace_block_idx in range(trace_curr_idx + 1, len(self._trace)):
+        for trace_block_idx in range(trace_curr_idx, len(self._trace)):
             trace_block = state.project.factory.block(self._translate_trace_addr(self._trace[trace_block_idx]))
             trace_block_last_insn = trace_block.capstone.insns[-1]
             for insn_type in control_flow_insn_types:
                 if trace_block_last_insn.group(insn_type):
-                    big_block_end_addr = trace_block.addr + trace_block.size
-                    break
+                    # Found first block in trace ending in a control flow instruction. Verify it matches the end of big
+                    # block according to VEX lifter
+                    big_block_end_addr = trace_block.addr + trace_block.size - 1
+                    if angr_big_block_end_addr != big_block_end_addr:
+                        # End does not match. Treat as trace desync.
+                        return False
+                    else:
+                        break
 
-        if big_block_end_addr is None:
+            if big_block_end_addr is not None:
+                break
+        else:
             # Failed to find end of the big basic block in trace. Treat as trace desync.
             return False
 
@@ -659,9 +751,9 @@ class Tracer(ExplorationTechnique):
         # - There is no control flow instruction between big_block_start_addr and big_block_end_addr
         # - There is no control flow instruction between big_block_start_addr and state_desync_block_addr
         # - state_desync_block_addr is definitely executed after big_block_start_addr
-        # So it's enough to check if state_desync_block_addr is less than big_block_end_addr to ensure that it
+        # So it's enough to check if desyncing block's address is less than big_block_end_addr to ensure that it
         # is part of the big block
-        return state_desync_block_addr < big_block_end_addr
+        return state.history.recent_bbl_addrs[state_desync_block_idx] < big_block_end_addr
 
     def _analyze_misfollow(self, state, idx):
         angr_addr = state.addr
@@ -742,7 +834,7 @@ class Tracer(ExplorationTechnique):
         try:
             sync_idx = self._trace.index(addr_translated, idx)
         except ValueError:
-            l.error("Trying to synchronize at %#x (%#x) but it does not appear in the trace?")
+            l.error("Trying to synchronize at %#x (%#x) but it does not appear in the trace?", addr_translated, addr)
             return False
 
         state.globals['sync_idx'] = sync_idx
@@ -777,7 +869,10 @@ class Tracer(ExplorationTechnique):
         # first check: are we just executing user-controlled code?
         if not state.ip.symbolic and state.mem[state.ip].char.resolved.symbolic:
             l.debug("executing input-related code")
-            return state
+            return state, state
+        # in case we can't unwind, we return the state itself
+        if state.addr == crash_addr:
+            return state, state
 
         state = state.copy()
         state.options.add(sim_options.COPY_STATES)
@@ -813,7 +908,7 @@ class Tracer(ExplorationTechnique):
         if inst_cnt == 0:
             insts = 0
         elif crash_addr in inst_addrs:
-            insts = inst_addrs.index(crash_addr) + 1
+            insts = inst_addrs.index(crash_addr)
         else:
             insts = inst_cnt - 1
 
@@ -834,14 +929,17 @@ class Tracer(ExplorationTechnique):
         l.debug("reconstraining... ")
         state.preconstrainer.reconstrain()
 
-        # now remove our breakpoints since other people might not want them
-        state.inspect.remove_breakpoint("address_concretization", bp1)
-        state.inspect.remove_breakpoint("address_concretization", bp2)
-
         l.debug("final step...")
-        succs = state.step()
+        succs = state.step(num_inst=1)
+
         successors = succs.flat_successors + succs.unconstrained_successors
         crash_state = successors[0]
+
+        # now remove our breakpoints since other people might not want them
+        for s in [last_state, crash_state]:
+            s.inspect.remove_breakpoint("address_concretization", bp1)
+            s.inspect.remove_breakpoint("address_concretization", bp2)
+
         return last_state, crash_state
 
     # the below are utility functions for crash windup
