@@ -10,7 +10,7 @@ if TYPE_CHECKING:
 class StructuredCodeManager(KnowledgeBasePlugin):
     def __init__(self, kb):
         self._kb = kb  # type: KnowledgeBase
-        self._codegens = {}
+        self.codegens = {}
 
     def _normalize_key(self, item):
         if type(item) is not tuple:
@@ -20,20 +20,21 @@ class StructuredCodeManager(KnowledgeBasePlugin):
         return item
 
     def __getitem__(self, item) -> 'StructuredCodeGenerator':
-        return self._codegens[self._normalize_key(item)]
+        return self.codegens[self._normalize_key(item)]
 
     def __setitem__(self, key, value):
-        self._codegens[self._normalize_key(key)] = value
+        self.codegens[self._normalize_key(key)] = value
 
     def __contains__(self, key):
-        return self._normalize_key(key) in self._codegens
+        return self._normalize_key(key) in self.codegens
 
     def available_flavors(self, item):
         if type(item) is str:
             item = self._kb.labels.lookup(item)
-        return [flavor for func, flavor in self._codegens if func == item]
+        return [flavor for func, flavor in self.codegens if func == item]
 
     def copy(self):
         raise NotImplementedError
 
+
 KnowledgeBasePlugin.register_default('structured_code', StructuredCodeManager)
